# This module houses tranformation functions applicable to arc-world grid & layers
# NOTE: For now, they are simple transforms that wrap numpy's standard array manipulations
import numpy as np


# Translate function wraps numpy's roll function to shift the grid in a given direction
# by x and y units while allowing for wrap-around and no-wrap-around options
def translate(grid: np.array, x: int, y: int, wrap: bool = True) -> np.array:
    """Translate the grid by x and y units"""
    if wrap:
        return np.roll(grid, (x, y), axis=(1, 0))
    else:
        # Since numpy's roll function doesn't support no-wrap-around,
        # the array has to be manually shifted by x and y units and padded with zeros
        shifted = np.zeros_like(grid)

        y1, y2 = max(0, y), min(grid.shape[0], grid.shape[0] + y)
        x1, x2 = max(0, x), min(grid.shape[1], grid.shape[1] + x)

        y1_, y2_ = max(0, -y), min(grid.shape[0], grid.shape[0] - y)
        x1_, x2_ = max(0, -x), min(grid.shape[1], grid.shape[1] - x)

        shifted[y1:y2, x1:x2] = grid[y1_:y2_, x1_:x2_]

        return shifted


def flip(grid: np.array, axis: int) -> np.array:
    """Flip the grid along the given axis"""
    return np.flip(grid, axis)


def rotate(grid: np.array, k: int) -> np.array:
    """Rotate the grid by 90 degrees k times"""
    return np.rot90(grid, k)


# These are grid level transformations since they have to apply to all layers in the grid
# They're right now left as identity wrappers around numpy's functions and can be extended if needed
def pad(grid: np.array, pad_width: int, mode: str = "constant") -> np.array:
    """Pad the grid with zeros"""
    return np.pad(grid, pad_width, mode)


def trim_zeros(grid: np.array) -> np.array:
    """Trim zeros from the grid"""
    return np.trim_zeros(grid)


def tile(grid: np.array, reps: tuple) -> np.array:
    """Tile the grid"""
    return np.tile(grid, reps)
