import random
from functools import partial

import numpy as np

from .actions import flip, rotate, tile, translate, trim_zeros


class Grid:
    """Grid represents a state in the arc-world as a 2D grid. This grid in turn has one or more stacked layers that represent objects in the world"""

    def __init__(self, num_rows: int, num_cols: int, num_objects: int = 1):
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.num_objects = num_objects

        # The zeroth layer is the base layer that only has zeros
        self.layers = [np.zeros((self.num_rows, self.num_cols), dtype=int)]
        self.array = self.layers[0]

        # Create objects in the grid depending on the number of objects
        for i in range(num_objects):
            layer = self._create_layer(i + 1)
            if layer is not None:
                self.layers.append(layer)

    def __repr__(self):
        return f"Grid(num_rows={self.num_rows}, num_cols={self.num_cols}, min_rows={self.min_rows}, min_cols={self.min_cols}, max_rows={self.max_rows}, max_cols={self.max_cols})"

    def __str__(self):
        return f"Grid: {self.num_rows}x{self.num_cols}"

    def _create_layer(self, color):
        # Create a layer for the object
        layer = np.zeros((self.num_rows, self.num_cols), dtype=int)
        attempts = 0
        while attempts < 5:
            start_row = random.randint(0, self.num_rows - 1)
            start_col = random.randint(0, self.num_cols - 1)
            if np.all(
                self.array[start_row - 2 : start_row + 3, start_col - 2 : start_col + 3]
                == 0
            ):
                break
            attempts += 1

        # If no valid starting point, skip creating the object
        if attempts == 5:
            return

        # Else, create the object
        layer[start_row, start_col] = color

        def add_neighbors(row, col, object_size, probability):
            for dr, dc in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
                new_row, new_col = row + dr, col + dc
                if (
                    0 <= new_row < self.num_rows
                    and 0 <= new_col < self.num_cols
                    and layer[new_row, new_col] == 0
                ):
                    adjusted_probability = probability * np.exp(-object_size / 10)
                    if random.random() < adjusted_probability:
                        layer[new_row, new_col] = color
                        object_size += 1
                        add_neighbors(new_row, new_col, object_size, probability * 0.5)

        add_neighbors(start_row, start_col, 1, 0.5)

        # Update the grid with the new layer
        self._update_grid(layer)

        return layer

    def _update_grid(self, layer):
        # Add non-zero elements from the layer to the grid
        self.array = np.maximum(self.array, layer)


class ArcWorld:
    """ArcWorld objects represent a world in the arcverse that is identified by behavior of objects within it represented as a series
    of transformation functions on different objects in the world"""

    def __init__(
        self,
        min_rows: int = 2,
        min_cols: int = 2,
        max_rows: int = 30,
        max_cols: int = 30,
    ):
        # Bounds for any grid in this world
        self.min_rows = min_rows
        self.min_cols = min_cols
        self.max_rows = max_rows
        self.max_cols = max_cols
        self.transforms = self.get_transform()
        self.puzzle = self._create_puzzle()

    def __repr__(self):
        return f"ArcWorld(min_rows={self.min_rows}, min_cols={self.min_cols}, max_rows={self.max_rows}, max_cols={self.max_cols})"

    def __str__(self):
        return f"ArcWorld: {self.min_rows}x{self.min_cols} to {self.max_rows}x{self.max_cols}"

    @property
    def min_rows(self):
        return self._min_rows

    @min_rows.setter
    def min_rows(self, value):
        if value < 2:
            raise ValueError("min_rows must be at least 2")
        self._min_rows = value

    @property
    def min_cols(self):
        return self._min_cols

    @min_cols.setter
    def min_cols(self, value):
        if value < 2:
            raise ValueError("min_cols must be at least 2")
        self._min_cols = value

    @property
    def max_rows(self):
        return self._max_rows

    @max_rows.setter
    def max_rows(self, value):
        if value <= self.min_rows or value > 30:
            raise ValueError(
                "max_rows must be greater than min_rows and less than or equal to 30"
            )
        self._max_rows = value

    @property
    def max_cols(self):
        return self._max_cols

    @max_cols.setter
    def max_cols(self, value):
        if value <= self.min_cols or value > 30:
            raise ValueError(
                "max_cols must be greater than min_cols and less than or equal to 30"
            )
        self._max_cols = value

    def get_transform(self):
        # For now, randomly select a fixed sequence of transformations only on the grid.
        transforms = []

        # Apply translations 70% of the time
        if random.random() < 0.5:
            # First randomly select a translation. Be strongly biased towards simple shifts
            if random.random() < 0.5:
                x_shift = random.randint(-3, 3)
            else:
                x_shift = 0
            # If x_shift is more than 1, then y_shift range is limited to 1
            if x_shift != 0 and random.random() < 0.5:
                y_shift = 0
            else:
                y_shift = (
                    random.randint(-1, 1) if abs(x_shift) > 1 else random.randint(-3, 3)
                )
            # Wrap 25% of the time
            wrap = random.random() < 0.5

            # Next, choose with some probability whether to translate, flip, rotate or tile

            # Translate
            transforms.append(partial(translate, x=x_shift, y=y_shift, wrap=wrap))

        # If translation is not applied, then apply flip, rotate or tile
        else:
            # Flip
            if random.random() < 0.5:
                axis = random.choice([0, 1])
                transforms.append(partial(flip, axis=axis))

            # Rotate
            elif random.random() < 0.5:
                k = random.randint(1, 3)
                transforms.append(partial(rotate, k=k))

        # If there are no transformations then tile at most 2x2
        if not transforms:
            transforms.append(
                partial(tile, reps=(random.randint(1, 2), random.randint(1, 2)))
            )

        # Finally if there are no transformations, simply crop the grid by trimming zeros
        if not transforms or random.random() < 0.1:
            transforms.append(trim_zeros)

        return transforms

    def sample(self, n: int = 1):
        # Give the transformation function, generate n examples of the arc-world

        samples = []
        for _ in range(n):
            # Generate a random size grid within max and min bounds
            num_rows = random.randint(self.min_rows, self.max_rows)
            num_cols = random.randint(
                max(self.min_cols, round(num_rows * 0.6)),
                min(self.max_cols, round(num_rows * 1.4)),
            )
            num_objects = random.randint(1, 1)
            grid = Grid(num_rows, num_cols, num_objects)
            startarr = grid.array
            nextarr = np.copy(startarr)
            # Apply the transformations to the grid array
            for transform in self.transforms:
                nextarr = transform(nextarr)

            samples.append((startarr, nextarr))

        return samples

    def _create_puzzle(self):
        # Create 5 training examples and one test example and save it in arc-puzzle format
        puzzle = {"train": [], "test": []}

        train_samples = self.sample(5)

        for sample in train_samples:
            puzzle["train"].append(
                {"input": sample[0].tolist(), "output": sample[1].tolist()}
            )

        test_sample = self.sample(2)

        for sample in test_sample:
            puzzle["test"].append(
                {"input": sample[0].tolist(), "output": sample[1].tolist()}
            )

        return puzzle

    def get_puzzle(self):
        return self.puzzle
